//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      factor.h: --- classes related to rgsbackmodel task
//
//      DESCRIPTION:
//             Calculates the factors needed by rgs background model
//
//      CHANGE HISTORY:
//              AI, 2004-06-24: 1.0.0: - first version
//
//      AUTHOR:
//              A. Ibarra (AI), Thu Jun 24 15:31:40 GMT 2004
//
//

#include "Factor.h"


typedef pair<const double, double> Pair;


Factor::Factor()
{

}


void Factor::runEvselect()
{

  std::string sasdir;
  if (getenv("SAS_DIR"))
    sasdir = getenv("SAS_DIR");
  else
    errstr << "Environmental variable SAS_DIR not defined " << Sas::error("SAS_DIR variable not found");

  string table(stringParameter("table"));

  order = intParameter("order");

  DataSet*  tempDS = dataSetServer->open(table, Dal::Read);

  //Get TLMAX14 attribute
  Table* tempTab = tempDS->table("EVENTS");



  // Check if the event file has the correct number of channels
  if (stringParameter("binningunits") == "BETA")
  {
	  Column *betaCol = tempTab->column("BETA_CHANNEL");
	  if ( betaCol->attribute("TLMAX")->asInt() != 3400 )
	  {
		  errstr << "Input event file incompatible with the background templates. "
		  << "Input file has "<< betaCol->attribute("TLMAX")->asInt()
		  << " beta channels and it must have 3400. "
		  << Sas::error("incompatible event file");
	  }
  }
  else
  {
	  Column *betaCol = tempTab->column("MLAMBDA_CHANNEL");
	  if ( betaCol->attribute("TLMAX")->asInt() != 3600 )
	  {
		  errstr << "Input event file incompatible with the background templates. "
		  << "Input file has "<< betaCol->attribute("TLMAX")->asInt()
		  << " beta channels and it must have 3600. "
		  << Sas::error("incompatible event file");
	  }

  }
  //Get TLMAX14 attribute


  /*
   * Check if there are data in CCD9. SOC-SPR-2600
   */
  ColumnData * CCDNRColData = tempTab->column("CCDNR")->data();
  int8 * ccdDat = CCDNRColData->int8Data();

  int ccd9Counter = 0;
  for (unsigned int i = 0 ; i < tempTab->rows() ; ++i)
    {
      if (ccdDat[i] == 9 )
	++ccd9Counter;
    }

  delete CCDNRColData;

  if (ccd9Counter == 0)
    errstr << "Background light curve cannot be computed "
	   << Sas::error("No CCD 9 data");


  if (tempDS->attribute("INSTRUME")->asString() == "RGS1")
    rgsunit = "RGS1";
  else if (tempDS->attribute("INSTRUME")->asString() == "RGS2")
       rgsunit = "RGS2";
  else
    errstr << "Instrument keyword RGS1 or RGS2 not found " << Sas::error("attribute not found");

  table=table+":EVENTS";


  //Size of the Time bin.
  //SOC-SPR-2685. Change bin size value from 60 to 100.
  double binSize =100;


  Info::Task* evselect = task("evselect");
  ParameterInfo& evselectpars = *evselect->parameters();
  evselectpars.revertToDefault();
  evselectpars["table"]->setValue(table);
  evselectpars["keepfilteroutput"]->setValue(false);
  evselectpars["withfilteredset"]->setValue(true);
  evselectpars["withrateset"]->setValue(true);
  evselectpars["rateset"]->setValue("myRate.fits");
  evselectpars["makeratecolumn"]->setValue(true);
  evselectpars["maketimecolumn"]->setValue(true);
  evselectpars["timebinsize"]->setValue(binSize);
  evselectpars["maketimecolumn"]->setValue(true);
  evselectpars["expression"]->setValue("(CCDNR==9)&&(XDSP_CORR< -0.0003||XDSP_CORR > 0.0003)");
  evselectpars["writedss"]->setValue(true);

  evselect->exec(true);

  dataSetServer->close(tempDS);

}

std::vector<double> Factor::calHistogram()
{

  DataSet* eventFile = dataSetServer->open("myRate.fits", Dal::Read);

  Table* events = eventFile->table("RATE");

  ColumnData * rateCol = events->column("RATE")->data();

  real64 *rateDat =  rateCol->real64Data();

  delete rateCol;

  std::vector<double> hist(16);

  //Get the binning defined in the CCF file

  calServer.setInstrument(getRGSunit().id());

  RgsBackgroundServer *rbs = 0; rbs = calServer.getAtom(rbs);

  // Compatibility with the new background template models binned in lambda space.
   string binningUnits =  stringParameter("binningunits");

  rbs->units(binningUnits);
  rbs->order(order);
  rbs->pipsfFrac(realParameter("pdistincl")/100.0);

  unsigned nlevel = rbs->numberExtensions();


  // Set the order to 1 and find the corresponding number of levels
  //unsigned nlevel = rbs->order(order).numberExtensions();

  for (unsigned l = 0; l<nlevel; ++l)
    {
      bin.push_back(rbs->level(l).bckgndLevel());

      //Add the last bin
      if (l == (nlevel -1 ) )
    	  bin.push_back(100.00);
    }

  calServer.releaseAtom(rbs);


  for (unsigned j = 0 ; j < bin.size()-1 ; j++)
    for (unsigned i = 0; i <  events->column("RATE")->rows(); i++)
      if ( (rateDat[i] >= bin[j]) && (rateDat[i] < bin[j+1]) )
	hist[j] = hist[j]+1;


  double sumHist=0;

  for (unsigned k =0 ; k < hist.size() ; k++)
    {
      errstr << "Histogram values: Bin-> "   <<  setprecision(2) <<
	bin[k] << " Val-> " << hist[k] << Sas::message(Msg::App, Msg::Verbose);
      sumHist = sumHist + hist[k];
    }


  if ( sumHist == 0 )
    errstr << " Histogram values empty. " <<Sas::error("Error calculating the histogram");

  // Normalize the histogram (Calculate factors)

  double checkFactors = 0.0;

  for (unsigned k =0 ; k < hist.size() ; k++)
    {
      factors.push_back(hist[k]/sumHist);

      errstr << "Bin-> " <<  setprecision(2)  << bin[k]
	     << setprecision(6)
	     << " Factor-> " << factors[k] << Sas::message(Msg::App, Msg::Verbose);
      checkFactors = checkFactors + factors[k];
    }

  errstr << "Sum of factors[i] = " << checkFactors <<  Sas::message(Msg::App, Msg::Verbose);


  dataSetServer->close(eventFile);
  dataSetServer->remove("myRate.fits");

  return factors;
}


std::vector<double> Factor::getBins()
{
  return bin;
}


XmmInstrument& Factor::getRGSunit()
{
  if (rgsunit == "RGS1")
    return xmmPayLoad.rgs1;
  else
    return xmmPayLoad.rgs2;
}

unsigned int Factor::getOrder()
{
  return order;
}

