//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      model.cc: --- classes related to rgsbackmodel task
//
//      DESCRIPTION:
//             Calculates the rates for the new background model.
//             and store the output in a new fits file called "Model.fits"
//
//      CHANGE HISTORY:
//              AI, 2004-06-25: 1.0.0: - first version
//
//      AUTHOR:
//              A. Ibarra (AI), Fri Jun 25 10:13:55 GMT 2004
//
//

#include "Model.h"


Model::Model(std::vector<double> b, std::vector<double> f,
	     XmmInstrument &r,
	     unsigned int o) :
  bin(b),factors(f),rgsunit(r),order(o)
{
  exposureTimeCCF = 0;
  rgsunit = xmmPayLoad.rgs1;
  M = bin.size() -1;
}

Model::~Model()
{
}

void Model::rateCal()
{
 //Define the number of channels depending on the BETA or LAMBDA parameter
 if (stringParameter("binningunits") == "LAMBDA")
	 _Num = 3600;
 else
	 _Num = 3400;

 double rateCCF[_Num];
 double errRateCCF[_Num];


 std::string fileName;
 string  sStr;

 double modelCCF[_Num][M];
 double errModelCCF[_Num][M];

 // Get the counts for all the templates and calculate the
 // exposure time

 calServer.setInstrument(rgsunit.id());

 RgsBackgroundServer *rbs = 0; rbs = calServer.getAtom(rbs);

 // Set the order to 1 and find the corresponding number of levels
 // Compatibility with the new background template models binned in lambda space.
  string binningUnits =  stringParameter("binningunits");

 rbs->units(binningUnits);
 rbs->order(order);
 rbs->pipsfFrac(realParameter("pdistincl")/100.0);


 unsigned nlevel = rbs->order(order).numberExtensions();

 // Retrieve the values from the CCFs
 for (unsigned l = 0; l<nlevel; ++l)
   {
     double expos = rbs->level(l).exposure();
     vector<int16> channel = rbs->level(l).channel();
     vector<int> counts = rbs->level(l).counts();
     vector<int16> quality = rbs->level(l).quality();
     vector<float> areascal = rbs->level(l).areascal();
     vector<float> backscal = rbs->level(l).backscal();

     exposureTimeCCF = exposureTimeCCF + (expos*factors[l]);

     for (unsigned int n = 0 ; n < counts.size() ; n++)
       {
	 modelCCF[n][l] = counts[n]/(areascal[n]*expos);
	 errModelCCF[n][l] = sqrt((double)counts[n])/(areascal[n]*expos);
       }
   }

 calServer.releaseAtom(rbs);

 errstr << "Background model exposure time "
	<< exposureTimeCCF << Sas::message(Msg::App, Msg::Verbose);


 // Calculate de background rate.
 double countsCCF[_Num];
 double errCountsCCF[_Num];


 // Array initialization
 for (unsigned int i = 0 ; i < _Num ; i++) rateCCF[i]=0;

 for (unsigned int i = 0 ; i < _Num ; i++) errRateCCF[i]=0;

 /*
  * Create the final rate (and error) multiplying each model factor
  * by the corresponding factor.
  *
  * Correct algorithm to compute the error is:
  * Error = sqrt(w1^2+E1^2 + w2^2*E2^2 + ...)
  * SOC-SPR-2598
  */

 for (unsigned int i = 0 ; i < _Num ; i++)
   {
     for (unsigned int j = 0 ; j < M ; j++)
       {
    	 rateCCF[i] = rateCCF[i] + modelCCF[i][j]*factors[j];
    	 errRateCCF[i] = errRateCCF[i] + pow(errModelCCF[i][j],2)*pow(factors[j],2);
       }
     errRateCCF[i] = sqrt( errRateCCF[i]);
   }


 for (unsigned int i = 0 ; i < _Num ; i++)
   {
     countsCCF[i] = rateCCF[i]*exposureTimeCCF;
     errCountsCCF[i] = errCountsCCF[i]*exposureTimeCCF;
   }

 bkgModel(rateCCF,errRateCCF);

}



// Create the Background model FITS file, with all the keywords needed
// to be OGIP complaint
void Model::bkgModel(double *rateCCF, double *errRateCCF)
{
  //openParameters("rgsbkgmodel");
  string table(stringParameter("table"));

  string rgsbkgmodelset(stringParameter("backgroundmodelset"));

  DataSet*  tempDS = dataSetServer->open(table, Dal::Read);

  //Copy the model in a fit file OGIP compliant
  DataSet* newSet = dataSetServer->open(rgsbkgmodelset,DataSet::Create);


  Table* modTabSpec = newSet->addTable("SPECTRUM",_Num);

  //  modTabSpec->addAttributes(tempDS);
  //  *modTabSpec->addAttribute("EXPOSURE","Exposure Time of the Background model","real unit") = exposureTimeCCF;



  //Get the template that correspond to the highest factor
  double tempMax=0.0;


  //Get the AREASCAL, BACKSCAL and QUALITY from the highest factor

  errstr <<"Instrument: "<<Xmm::toString(rgsunit.id())<< Sas::message(Msg::App, Msg::Verbose);
  errstr <<"Order: "<< order <<Sas::message(Msg::App, Msg::Verbose);
  errstr <<"pdistincl: "<< realParameter("pdistincl") <<Sas::message(Msg::App, Msg::Verbose);

  calServer.setInstrument(rgsunit.id());

  // Call access to the ccf files

  RgsBackgroundServer *rbs = 0;
  rbs = calServer.getAtom(rbs);

  // Compatibility with the new background template models binned in lambda space.
  string binningUnits =  stringParameter("binningunits");

  rbs->units(binningUnits);
  rbs->order(order);
  rbs->pipsfFrac(realParameter("pdistincl")/100.0);

  vector<int16> channel;
  vector<int16> quality;
  vector<float> areascal;
  vector<float> backscal;

  double bestBin=0.0;

  //factor.size() and nlevel has the same number of bins.
  //Select the bin with the hightest value.
  for (unsigned int i = 0 ; i < factors.size() ; i++)
    tempMax = max(tempMax,factors[i]);

  for (unsigned int i = 0 ; i < factors.size() ; i++)
    {
      if (tempMax == factors[i])
	{
	  bestBin = bin[i];
	  errstr<< "maximum count rate " <<  factors[i] << endl;
	  errstr<< "Bin that correspond to the maximum count rate " <<  rbs->level(i).bckgndLevel() << endl;
	  channel = rbs->level(i).channel();
	  quality = rbs->level(i).quality();
	  areascal= rbs->level(i).areascal();
	  backscal = rbs->level(i).backscal();
	  errstr << "AREASCAL, BACKSCAL and QUALITY taken from bin: " << bin[i]
		 <<Sas::message(Msg::App, Msg::Verbose);
	}
    }

  calServer.releaseAtom(rbs);

  Column *colChannel = modTabSpec->addColumn("CHANNEL",Column::Int16,"dispersion channel");
  ColumnData *dChannel = colChannel->data();
  int16 *i16Channel = dChannel->int16Data();
  for (unsigned long i = 0; i < _Num; ++i)
    i16Channel[i] = channel[i];
  delete dChannel;


  Column *colRateCCF = modTabSpec->addColumn("RATE",Column::Real32,"real32 data","counts/sec");
  ColumnData *dCCF = colRateCCF->data();
  real32 *r32CCF = dCCF->real32Data();
  for (unsigned long i = 0; i < _Num; ++i)
    r32CCF[i] = rateCCF[i];

  delete dCCF;

  Column *colErrRateCCF = modTabSpec->addColumn("STAT_ERR",Column::Real32,"real32 data","counts/sec");
  ColumnData *errRateDatCCF = colErrRateCCF->data();
  real32 *err32CCF = errRateDatCCF->real32Data();
  for (unsigned long i = 0; i < _Num; ++i)
    err32CCF[i] = errRateCCF[i];

  delete errRateDatCCF ;


  Column *colQuality = modTabSpec->addColumn("QUALITY",Column::Int16,"quality flag");
  ColumnData *dQuality = colQuality->data();
  int16 *i16Quality = dQuality->int16Data();
  for (unsigned long i = 0; i < _Num; ++i)
    i16Quality[i] = quality[i];
  delete dQuality;


  //AREASCAL values need to be checked before beeing written in the
  //backgound model file.
  // If the areascal is smaller than 0.6 set to 1.0
  for (unsigned long i = 0; i < _Num; ++i)
    if ((areascal[i] <= 0.6) && (areascal[i] != 0.0) )
      areascal[i] = 1.0;


  Column *colAreascal = modTabSpec->addColumn("AREASCAL",Column::Real32,"area scaling factor");
  ColumnData *dAreascal = colAreascal->data();
  real32 *r32Areascal = dAreascal->real32Data();
  for (unsigned long i = 0; i < _Num; ++i)
    r32Areascal[i] = areascal[i];
  delete dAreascal;



  Column *colBackscal = modTabSpec->addColumn("BACKSCAL",Column::Real32,"background scaling factor");
  ColumnData *dBackscal = colBackscal->data();
  real32 *r32Backscal = dBackscal->real32Data();
  for (unsigned long i = 0; i < _Num; ++i)
    r32Backscal[i] = backscal[i];
  delete dBackscal;



  Table* modTabFac = newSet->addTable("FACTORS",M);
  modTabFac->addAttributes(newSet);

  Column *colFactors = modTabFac->addColumn("Factors",Column::Real64,"real64 data","%");
  ColumnData *dFactors = colFactors->data();
  real64 *factors64 = dFactors->real64Data();
  for (unsigned long i = 0; i < M; ++i)
    factors64[i] = factors[i];
  delete dFactors;



  /*-----------------Add Attributes---------------------*/

  // Primary header keywords.
  std::string objName;
  objName = "Bkg model for obsid " + tempDS->attribute("OBS_ID")->asString();

  newSet->addAttribute( tempDS->attribute("TELESCOP") );
  newSet->addAttribute( tempDS->attribute("INSTRUME") );
  newSet->addAttribute( tempDS->attribute("FILTER") );
  *newSet->addAttribute("OBJECT","Name of target") = objName;
  newSet->addAttribute( tempDS->attribute("OBS_ID") );
  newSet->addAttribute( tempDS->attribute("EXP_ID") );


  float lineSep;
  struct MiscDataServer *misc = calServer.getAtom(misc);
  lineSep = misc->get("GRAT_LINE_DENS");
  /*
    rgsAlpha = (misc->get("LINCOORD:LINCOORD%G_RY")
    + misc->get("INCIDENCE_ANGLE")
    + (DalVecE(table->column("DELTA_DISP"))[i]/60)
    ) * M_PI/180;
  */

  // Keywords to include in the spectrum headers.
  modTabSpec->addAttribute( tempDS->attribute("TELESCOP") );
  modTabSpec->addAttribute( tempDS->attribute("INSTRUME") );
  modTabSpec->addAttribute( tempDS->attribute("FILTER") );
  *modTabSpec->addAttribute("OBJECT","Name of target") = objName;
  modTabSpec->addAttribute( tempDS->attribute("OBS_ID") );
  modTabSpec->addAttribute( tempDS->attribute("EXP_ID") );
  *modTabSpec->addAttribute("HDUCLASS","format conforms to OGIP standard","")="OGIP";
  *modTabSpec->addAttribute("HDUCLAS1","OGIP Memo OGIP/92-007","")="SPECTRUM";
  *modTabSpec->addAttribute("HDUCLAS2","what combination of source and background","")="BKG";
  *modTabSpec->addAttribute("HDUCLAS3","which data representation","")="RATE";
  *modTabSpec->addAttribute("HDUVERS","version of format specification","")="1.2.0";
  *modTabSpec->addAttribute("CHANTYPE","which channel calibration applies","")="PI";
  *modTabSpec->addAttribute("DETCHANS","total number of detector channels","")=(int)_Num;
  *modTabSpec->addAttribute("SYS_ERR","no systematic error specified","")=0;
  *modTabSpec->addAttribute("GROUPING","no channel grouping specified","")=0;
  *modTabSpec->addAttribute("CORRSCAL","correlation scale factor","")=1;
  *modTabSpec->addAttribute("CORRFILE","associated correlation file","")="none";
  *modTabSpec->addAttribute("RESPFILE","associated redistribution matrix","")="none";
  *modTabSpec->addAttribute("ANCRFILE","associated ancillary response file","")="none";
  *modTabSpec->addAttribute("BACKFILE","associated background spectrum","")="none";
  *modTabSpec->addAttribute("SOURCEID","source extracted","SRCLIST:INDEX")=2;
  *modTabSpec->addAttribute("RFLORDER","reflection order extracted","")=(int)order;
  *modTabSpec->addAttribute("LINE_SEP","grating line spacing","Angstroms")=lineSep;
  *modTabSpec->addAttribute("RGSALPHA","grating incidence angle","rad")=2.57097240537405E-02;
  *modTabSpec->addAttribute("EXPOSURE","maximum effective integration time","s")=exposureTimeCCF;

  *colChannel->addAttribute("TCTYP","channel axis type") = "";

  if (stringParameter("binningunits") == "LAMBDA")
  {
	  *colChannel->addAttribute("TCUNI","channel axis units (-Angstrom*order)") = "Angstrom";
	  *colChannel->addAttribute("TCRPX","reference channel number") = 1;
	  *colChannel->addAttribute("TCRVL","center of reference channel")= 4.005;
	  *colChannel->addAttribute("TCDLT","uniform channel width") = 0.01;
  }
  else
  {
	  *colChannel->addAttribute("TCUNI","channel axis units") = "rad";
	  *colChannel->addAttribute("TCRPX","reference channel number") = 1;
	  *colChannel->addAttribute("TCRVL","center of reference channel")= 3.57852391898632E-02;
	  *colChannel->addAttribute("TCDLT","uniform channel width") = 1.20799995784182E-05;
  }

  *colChannel->addAttribute("TLMIN") = 1;
  *colChannel->addAttribute("TLMAX") = (int)_Num;



  dataSetServer->close(newSet);
  dataSetServer->close(tempDS);
  //closeParameters();
}
